(* 
 * Implementation for 
 *   Semantic Checking
 * Author: Lixing Dong
 * Email: cee.xingweimian@gmail.com
 *)

open Ast

module StringMap = Map.Make(String)

module Semantic =
struct


type func_record =
{
    return_tp : cltype;
    (*paramter type list*)
    tplist : cltype list;
}

type env = 
{
    mutable funcmap : func_record StringMap.t;
    mutable table_list : (cltype StringMap.t) list;
    mutable objmap :  (cltype StringMap.t) StringMap.t;
}


(* variable of type env *)
let ienv =
{
    funcmap = StringMap.empty;
    table_list = [StringMap.empty];
    objmap = StringMap.empty;
}


type tpli =
{
    line : int;
    tp : cltype;
}


(* whether no error exists *)
let checking_succeed = ref true

(*whether current node is in loop*)
let nLoop = ref 0
let inFunc = ref 0
let return = ref 0
let tpFunc = ref ""
let currFuncName = ref ""

let bHasBreak = ref false

let bInLoop () =
    if !nLoop > 0 then
        true
    else
        false

let bInFunc () =
    if !inFunc > 0 then
        true
    else
        false

let returned () =
    if !return > 0 then
        true
    else
        false


let rec string_of_cltype (t: cltype) =
    match t with
    CLTypeOfInt -> "int"
    | CLTypeOfReal -> "float"
    | CLTypeOfObjID id -> id
    | CLTypeOfBoolean -> "bool"
    | CLTypeOfArray (tp, le)-> "array of " ^ string_of_cltype tp
    | CLTypeOfUndeclare -> "undeclare"
    | CLTypeOfVoid -> "void"



(* find the type of string s *)
let fold_fun (s, tp) (map: 'a StringMap.t) =
    match tp with
    CLTypeOfUndeclare -> 
        (try 
            let tp = StringMap.find s map in
            (s, tp)
        with Not_found -> 
            (s, CLTypeOfUndeclare))
    | _ ->
        (s, tp)


let check_variable (s: string) (li: int) =
    let (_, tpn) = 
        List.fold_left fold_fun (s, CLTypeOfUndeclare) ienv.table_list 
    in
    match tpn with
    CLTypeOfUndeclare ->
        let errormsg = "Semantic Error: Undeclared variable '" ^ s ^ "' at line " ^ (string_of_int li) in
        Printf.printf "%s\n" errormsg;
        checking_succeed := false;
        { tp = CLTypeOfUndeclare; line = li}
    | _ ->
        { tp = tpn; line = li}


let find_variable_in_local (s: string) =
    let local = List.hd ienv.table_list in
    StringMap.mem s local


let add2table (s: string) (tp: cltype) (li: int) = 
    if find_variable_in_local s then
        let errormsg = 
            "Semantic Error: Redefinition of variable '" ^ s ^ "' at line: " ^ string_of_int li ^ "\n" in
        Printf.printf "%s" errormsg;
checking_succeed := false
    else
        let local = List.hd ienv.table_list in
        let local = StringMap.add s tp local in
        ienv.table_list <- local :: List.tl ienv.table_list


let check_int (e1: tpli)  =
    match e1.tp with
    CLTypeOfInt -> ()
    | CLTypeOfUndeclare -> ()
    | _ ->
        let errormsg = "Semantic Error: expected type 'int' but have type '" ^ string_of_cltype e1.tp ^ "' at line " ^ (string_of_int e1.line) in
        Printf.printf "%s\n" errormsg;
        checking_succeed := false


let check_bool (e1: tpli)  = 
    match e1.tp with
    CLTypeOfBoolean -> 
        ()
    | CLTypeOfUndeclare -> ()
    | _ ->
        let errormsg = "Semantic Error: expected type 'bool' but have type '" ^ string_of_cltype e1.tp ^ "' at line " ^ (string_of_int e1.line) in
        Printf.printf "%s\n" errormsg;
        checking_succeed := false


let check_rgb (e1: tpli) (e2: tpli) (e3: tpli)  = 
    check_int e1;
    check_int e2;
    check_int e3


(*
 *
 * check rules
 *
 *)

 let rec check (lst: stc list) = 
ienv.objmap <- StringMap.add "int" StringMap.empty ienv.objmap;
ienv.objmap <- StringMap.add "bool" StringMap.empty ienv.objmap;
ienv.objmap <- StringMap.add "void" StringMap.empty ienv.objmap;
    match lst with
    | [] -> ()
    | head::[] -> check_stc head
    | head::tail -> 
        check_stc head;
        check tail


and check_stc (s: stc) = 
    match s with
    | StcOfObjDef o -> check_obj_dec o;
    | StcOfFuncDef f -> check_fun_dec f;
    | StcOfStmt s -> check_stmt s;


and check_cst_expr (e: cst_expr) (li: int) =
    match e with
    CstExprOfInt (s) -> {tp = CLTypeOfInt; line = li}
    | CstExprOfReal (s) -> {tp = CLTypeOfReal; line = li}
    | CstExprOfTrue
    | CstExprOfFalse -> {tp = CLTypeOfBoolean; line = li}

and check_elem_exist (tp: cltype) (r: identifier) (li: int)= 
match tp with
|CLTypeOfUndeclare -> CLTypeOfUndeclare
|_ -> check_elem_exist2 tp r li

and check_elem_exist2 (tp: cltype) (r: identifier) (li: int) = 
    match r with 
    | IdentifierOfSimple (id, li) -> check_elem_exist3 id tp li
	
    | IdentifierOfObjElement (l, r, li) -> check_elem_exist4 l tp li r
	

and check_elem_exist3 (id :string) (tp: cltype) (li: int) =
	try
        StringMap.find id (StringMap.find (string_of_cltype tp) ienv.objmap)
    	with Not_found ->
            let errormsg = "Semantic Error: Object '" ^ (string_of_cltype tp) ^ "' has no member called '" ^ id ^ "' at line " ^ (string_of_int li) in
            Printf.printf "%s\n" errormsg;
            checking_succeed := false;
            CLTypeOfUndeclare
and check_elem_exist4 (l :string) (tp: cltype) (li: int) (r:identifier) =
	try
        let parent_tp = StringMap.find l (StringMap.find (string_of_cltype tp) ienv.objmap) in 
	
	check_elem_exist parent_tp r li
    	with Not_found ->
            let errormsg = "Semantic Error: Object '" ^ (string_of_cltype tp) ^ "' has no member called '" ^ l ^ "' at line " ^ (string_of_int li) in
            Printf.printf "%s\n" errormsg;
            checking_succeed := false;
            CLTypeOfUndeclare

and check_identifier (id: identifier) (li: int) =
    match id with
    | IdentifierOfSimple (id, li) -> check_variable id li
    | IdentifierOfObjElement (l, r, li) -> 
	let parent_tp = (check_variable l li).tp in
	{ tp = (check_elem_exist parent_tp r li); line = li;}
        

and check_primary (p: primary) (li: int) =
    match p with
    | PrimaryOfIdentifier (id, li) -> 
        check_identifier id li
    | PrimaryOfArrayElement (id, e) ->
        let e =  (check_expr e) in
	check_int (e);
        let tmp = check_identifier id e.line in
	let li = tmp.line in
	match tmp.tp with
		|CLTypeOfArray (t, l) -> {tp = t; line = li}
		|_ ->
	let errormsg = "Semantic Error: expected type 'array', but have type'" ^ string_of_cltype tmp.tp ^ "' at line " ^ (string_of_int tmp.line) in
        Printf.printf "%s\n" errormsg;
        checking_succeed := false;
	{tp = CLTypeOfUndeclare; line = li}


and check_lvalue (l: lvalue)=
    match l with
    LvalueOfIdentifier (id, li) -> 
        check_identifier id li
    | LvalueOfArrayElement (id, e) ->
        let e =  (check_expr e) in
	check_int (e);
        let tmp = check_identifier id e.line in
	let li = tmp.line in
	match tmp.tp with
		|CLTypeOfArray (t, l) -> {tp = t; line = li}
		|_ ->
	let errormsg = "Semantic Error: expected type 'array', but have type'" ^ string_of_cltype tmp.tp ^ "' at line " ^ (string_of_int tmp.line) in
        Printf.printf "%s\n" errormsg;
        checking_succeed := false;
	{tp = CLTypeOfUndeclare; line = li}
		


and check_typematch (l: tpli) (v: tpli) = 
    if string_of_cltype l.tp = "undeclare" then
        l
    else if string_of_cltype v.tp = "undeclare" then
        v
    else
    if l.tp = v.tp then
        l
    else
        let errormsg = "Semantic Error: expected type '" ^ string_of_cltype l.tp ^ "' but have type'" ^ string_of_cltype v.tp ^ "' at line " ^ (string_of_int l.line) in
        Printf.printf "%s\n" errormsg;
        checking_succeed := false;
        l
        

and check_assign (l: tpli) (v: tpli)  = 
    check_typematch l v


and check_bop (e1: tpli) (e2: tpli) = 
    if string_of_cltype e1.tp = "int" then 
        check_typematch e1 e2
    else 
        let errormsg = "Semantic Error: no operation defined for type '" ^ string_of_cltype e1.tp ^ "' at line " ^ (string_of_int e1.line) in
        Printf.printf "%s\n" errormsg;
        checking_succeed := false;
        e1


and check_boolop (e1: tpli) (e2: tpli)  = 
    if string_of_cltype e1.tp = "int" then
	let e = {
        line = e1.line;
        tp = CLTypeOfBoolean;
    } in 
        check_typematch e1 e2; e
    
    else (let errormsg = "Semantic Error: no operation defined for type '" ^ string_of_cltype e1.tp ^ "' at line " ^ (string_of_int e1.line) in
            Printf.printf "%s\n" errormsg;
            checking_succeed := false;
            {
                line = e1.line;
                tp = CLTypeOfBoolean;
            });
    {
        line = e1.line;
        tp = CLTypeOfBoolean;
    }

and get_func_call_para_list (lst: expr list)=
match lst with
    | [] -> [CLTypeOfVoid]
    | head :: [] -> [(check_expr head).tp]
    | head :: tail ->
        let head = (check_expr head).tp in
        let tail = get_func_call_para_list tail in
        head :: tail


and check_func_call_para (id: string) (lst: expr list) (li: int)=
let lst = get_func_call_para_list lst in 
let errormsg = "Semantic Error: Function '" ^ id ^ "' argument do not match with definition at  line " ^ (string_of_int li) in
	try 
		match StringMap.find id ienv.funcmap with
			|lst ->{tp = CLTypeOfUndeclare; line = li}
			|_ -> Printf.printf "%s\n" errormsg;
            			checking_succeed := false;
				{tp = CLTypeOfUndeclare; line = li}
    	with Not_found ->
Printf.printf "%s\n" errormsg;
            checking_succeed := false;
            {tp = CLTypeOfUndeclare; line = li}


and check_func_return_type (id: string) (li: int)=
try
        (*Printf.printf "%d\n\n" (StringMap.cardinal ienv.global);*)
        {tp = (StringMap.find id ienv.funcmap).return_tp; line = li}
    with Not_found ->
            let errormsg = "Semantic Error: Undeclared function '" ^ id ^ "' at  line " ^ (string_of_int li) in
            Printf.printf "%s\n" errormsg;
            checking_succeed := false;
            {tp = CLTypeOfUndeclare; line = li}


and check_expr (e: expr) =
    match e with
    | ExprOfPrimary (p, li) -> check_primary p li
    | ExprOfCstExpr (e, li) -> check_cst_expr e li
    | ExprOfAssign (l, v) -> 
        let l = check_lvalue l in
        let v = check_expr v in
        check_assign l v
    | ExprOfPlus (e1, e2)
    | ExprOfMinus (e1, e2)
    | ExprOfTimes (e1, e2)
    | ExprOfDivide (e1, e2) ->
        let e1 = check_expr e1 in
        let e2 = check_expr e2 in
        check_bop e1 e2
    | ExprOfNeq (e1, e2)
    | ExprOfEq (e1, e2)
    | ExprOfLt (e1, e2)
    | ExprOfGt (e1, e2)
    | ExprOfLeq (e1, e2)
    | ExprOfGeq (e1, e2) ->
        let e1 = check_expr e1 in
        let e2 = check_expr e2 in
        check_boolop e1 e2
    | ExprOfUMinus (e1) -> 
        let e1 = check_expr e1 in
        check_int e1;
        e1
    | ExprOfNot e ->
        let e = check_expr e in
        check_bool e;
        e
    | ExprOfFunction (id, lst, li) ->
	ignore(check_func_call_para id lst li);
        check_func_return_type id li;


and check_expr_stmt (e: expr_stmt) =
    match e with
    ExprStmtOfExpr e -> ignore(check_expr e); ()


and check_cltype (t: cltype) =
    let st = string_of_cltype t in
	try
        StringMap.find st ienv.objmap;
	t
    with Not_found ->
            let errormsg = "Semantic Error: Undeclared type '" ^ st ^ "' " in
            Printf.printf "%s\n" errormsg;
            checking_succeed := false;
            CLTypeOfUndeclare

and get_int_from_cst_expr (e: cst_expr) =
match e with
    CstExprOfInt (s) -> s
    | CstExprOfReal (_)
    | CstExprOfTrue
    | CstExprOfFalse -> 0


and check_declarator (tp: cltype) (d: declarator) =
    (*"TODO"*)
    match d with
    DeclaratorOfID (s, li) ->
        add2table s tp li
    | DeclaratorOfArray (s, li) ->
	let tp = CLTypeOfArray(tp, 0) in
        add2table s tp li
    | DeclaratorOfFixedArray (s, e, li) ->
        let e = get_int_from_cst_expr e in
	let tp = CLTypeOfArray(tp, e) in
	let errormsg = "Semantic Error: Array index must be greater than 0 at line " ^ (string_of_int li) in	
	if e > 0 then
        add2table s tp li
	else 
            (Printf.printf "%s\n" errormsg;
            checking_succeed := false)


and check_declarator_list (tp: cltype) (declst: declarator list) =
    (*"TODO"*)
    match declst with
    [] -> raise (Failure("Empty Declarator List\n"))
    | head :: [] -> check_declarator tp head
    | head :: tail -> 
        check_declarator tp head;
        check_declarator_list tp tail


and check_var_dec (vc: var_dec) =
    match vc with
    VarDecOfTypeAndDeclaratorList (tp, declst) ->
        ignore(check_cltype tp);
        check_declarator_list tp declst


and assign_local_var (tp: cltype) (id: string) =
    (*"TODO"*)
    add2table id tp 0


and check_para (para: parameter) =
    match para with ParameterOfTypeAndID (tp, id) ->
    let tp = check_cltype tp in
    assign_local_var tp id;
    tp;


and check_paralist (paralist: parameter list) =
    match paralist with
    | [] -> [CLTypeOfVoid]
    | head::[] -> [check_para head]
    | head::tail -> 
        let head = check_para head in
        let tail = check_paralist tail in
        head :: tail


and assign_func (id: string) (tplist: cltype list) (tp : cltype) =
    (*"TODO"*)
   match  StringMap.mem id ienv.funcmap with
|false ->
    let item = 
    {
        return_tp = tp;
        tplist = (*TODO*)tplist;
    } in
    ienv.funcmap <- StringMap.add id item ienv.funcmap
|true -> let errormsg = 
            "Semantic Error: Redefinition of function '" ^ id ^ "'\n" in
        Printf.printf "%s" errormsg;
checking_succeed := false
	


and check_id_para (id_para: func_dec) (tp: cltype)=
    match id_para with FuncDeclaratorOfIDAndParameters (id, paralist) ->
    let tplist = check_paralist paralist in
    currFuncName := id;
    assign_func id tplist tp


and check_in_func_stc (stc: in_func_stc) = 
    match stc with
    InFuncStcOfStmt s -> check_stmt s;


and check_func_body (lst: in_func_stc list) =
    match lst with
    | [] -> ()
    | head::[] -> check_in_func_stc head
    | head::tail -> 
        check_in_func_stc head;
        check_func_body tail


and check_fun_dec (f: func_def) =
    match f with 
    FuncDefOfFuncDecAndStcList (id_para, tp, lst) ->
        ienv.table_list <- StringMap.empty :: ienv.table_list;
        check_id_para id_para tp;
        let tp = check_cltype tp in
		tpFunc := string_of_cltype tp;
		inFunc := 1;
                check_func_body lst;
	match not(returned ()) && (!tpFunc <> "void") with
		|true ->let errormsg = "Syntax Error: Function '" ^ !currFuncName ^ "' without return." ^ !tpFunc in
        		Printf.printf "%s\n" errormsg;
            		checking_succeed := false;
			ienv.table_list <- List.tl ienv.table_list;
			inFunc := 0;
			tpFunc := "";
			return := 0;
			currFuncName := ""
		|false -> ienv.table_list <- List.tl ienv.table_list;
			inFunc := 0;
			tpFunc := "";
			return := 0;
			currFuncName := ""

and check_in_obj_dec (s: in_obj_stc) = 
    match s with
	|InObjStcOfVarDec (vd) -> check_var_dec vd

and check_obj_body (lst: in_obj_stc list) =
    match lst with
    | [] -> ()
    | head::[] -> check_in_obj_dec head
    | head::tail -> check_in_obj_dec head;
	check_obj_body tail;



and check_obj_dec (o: obj_def) =
    match o with 
    | ObjDefOfIDAndStcList(id, lst) ->
		match  StringMap.mem id ienv.objmap with
		|false -> 
	ienv.table_list <- StringMap.empty :: ienv.table_list;
	let cltypeid = CLTypeOfObjID (id) in
        (*TODO: check duplicate*)
	check_obj_body lst;
	let local = List.hd ienv.table_list in 
		ienv.objmap <- StringMap.add id local ienv.objmap;
	ienv.table_list <- List.tl ienv.table_list
		|true ->  let errormsg = 
            "Semantic Error: Redefinition of object '" ^ id ^ "'\n" in
        Printf.printf "%s" errormsg;
checking_succeed := false



and check_draw_stmt (ds: draw_stmt) =
    match ds with
    | DrawStmtOfFD e
    | DrawStmtOfBK e
    | DrawStmtOfLF e
    | DrawStmtOfRT e
    | DrawStmtOfWd e -> check_int (check_expr e)
    | DrawStmtOfRGB (e1, e2, e3) ->
        let e1 = check_expr e1 in
        let e2 = check_expr e2 in
        let e3 = check_expr e3 in
        check_rgb e1 e2 e3
    | _ -> ()


and check_stmt (s: stmt) =
    match s with
    | StmtOfVarDec v -> check_var_dec v;
    | StmtOfExprStmt e -> check_expr_stmt e
    | StmtOfSelectionStmt s -> check_selection_stmt s
    | StmtOfIterationStmt is -> check_iteration_stmt is
    | StmtOfJumpStmt js -> check_jump_stmt js
    | StmtOfDrawStmt ds -> check_draw_stmt ds


and check_jump_stmt (s: jump_stmt) =
    match s with
    JumpStmtOfGOON li->
        if not(bInLoop ()) then
	    let errormsg = "Syntax Error: Goon statement must be in loop at line " ^ (string_of_int li) in
        	Printf.printf "%s\n" errormsg;
            checking_succeed := false
    | JumpStmtOfBRK li-> 
        (if not(bInLoop ()) then
	    let errormsg = "Syntax Error: Break statement must be in loop at line " ^ (string_of_int li) in
        	Printf.printf "%s\n" errormsg;
            checking_succeed := false);
	bHasBreak := true
    | JumpStmtOfReturn li -> 
        if not(bInFunc ()) then
	    let errormsg = "Syntax Error: Return statement must be in function at line " ^ (string_of_int li) in
        	Printf.printf "%s\n" errormsg;
            checking_succeed := false
	else if (!tpFunc <> "void") then
	    let errormsg = "Syntax Error: Return type should be '" ^ !tpFunc ^ "' but is 'void' at line " ^ (string_of_int li) in
        	Printf.printf "%s\n" errormsg;
            checking_succeed := false
	else return := !return + 1;
    | JumpStmtOfReturnExpr e ->
	let e = check_expr e in
	if not(bInFunc ()) then
	    let errormsg = "Syntax Error: Return statement must be in function at line " ^ (string_of_int e.line) in
        	Printf.printf "%s\n" errormsg;
            checking_succeed := false
	else 
		
	    if (!tpFunc <> string_of_cltype e.tp) then
	    let errormsg = "Syntax Error: Return type should be '" ^ !tpFunc ^ "' but is '" ^ string_of_cltype e.tp ^ "' at line " ^ (string_of_int e.line) in
        	Printf.printf "%s\n" errormsg;
            checking_succeed := false
        else return := !return + 1;


and check_iteration_stmt (s: iteration_stmt) =
    nLoop := !nLoop + 1;
    ienv.table_list <- StringMap.empty :: ienv.table_list;
    match s with
    | IterationStmtOfExprAndStmtList (e, lst) ->
        check_stmt_list lst;
        let e = check_expr e in
        check_int e
            
    | IterationStmtOfStmtList (lst) ->

        (*"TODO"*)check_stmt_list lst;
	(
		match not !bHasBreak with 
		|true ->Printf.printf "Semantic Error: Infinite Loop Without Break\n";
        			checking_succeed := false
        | _ -> ()

	);
        
        bHasBreak := false
	;
    ienv.table_list <- List.tl ienv.table_list;
    nLoop := !nLoop - 1


and check_selection_stmt (s: selection_stmt) =
    match s with
    | SelectionStmtOfExprAndStmtList(e, lst) ->
        check_stmt_list lst;
        let e = check_expr e in
        check_bool e
            
    | SelectionStmtOfExprAndStmtList2(e, lst, elst) ->
        check_stmt_list lst;
        check_stmt_list elst;
        let e = check_expr e in
        check_bool e


and check_stmt_list (lst: stmt list) =
    match lst with
    | [] -> ()
    | head :: [] -> check_stmt head;
    | head :: tail ->
        check_stmt head;
        check_stmt_list tail



end

